@@ -35,6 +35,7 @@ namespace Tomboy
 			list.Add (Gdk.Atom.Intern ("_NETSCAPE_URL", false), 0, 1);
 			
 			KeyPressEvent += KeyPressed;
+			ButtonPressEvent += ButtonPressed;
 		}
 
 		public static int DefaultMargin
@@ -92,32 +93,46 @@ namespace Tomboy
 
 			if (has_url) {
 				UriList uri_list = new UriList (selection_data);
-				StringBuilder insert = new StringBuilder ();
+				bool more_than_one = false;
+				
+				// Place the cursor in the position where the uri was
+				// dropped, adjusting x,y by the TextView's VisibleRect.
+				Gdk.Rectangle rect = VisibleRect;
+				int adjustedX = x + rect.X;
+				int adjustedY = y + rect.Y;
+				Gtk.TextIter cursor = GetIterAtLocation (adjustedX, adjustedY);
+				Buffer.PlaceCursor (cursor);
+				
+				Gtk.TextTag link_tag = Buffer.TagTable.Lookup ("link:url");
 
 				foreach (Uri uri in uri_list) {
 					Logger.Log ("Got Dropped URI: {0}", uri);
-
-					// FIXME: The space here is a hack
-					// around a bug in the URL Regex which
-					// matches across newlines.
-					if (insert.Length > 0)
-						insert.Append (" \n");
-
-					if (uri.IsFile) 
-						insert.Append (uri.LocalPath);
+					string insert;
+					if (uri.IsFile)
+						insert = uri.LocalPath;
 					else
-						insert.Append (uri.ToString ());
+						insert = uri.ToString ();
+					
+					if (insert == null || insert.Trim () == String.Empty)
+						continue;
+					
+					if (more_than_one) {
+						cursor = Buffer.GetIterAtMark (Buffer.InsertMark);
+						
+						// FIXME: The space here is a hack
+						// around a bug in the URL Regex which
+						// matches across newlines.
+						if (cursor.LineOffset == 0)
+							Buffer.Insert (ref cursor, " \n");
+						else
+							Buffer.Insert (ref cursor, ", ");
+					}
+					
+					Buffer.InsertWithTags (ref cursor, insert, link_tag);
+					more_than_one = true;
 				}
 
-				if (insert.Length > 0) {
-					Gtk.TextIter insert_iter = 
-						Buffer.GetIterAtMark (Buffer.InsertMark);
-					Buffer.InsertWithTags (ref insert_iter,
-							       insert.ToString (),
-							       Buffer.TagTable.Lookup ("link:url"));
-				}
-
-				Gtk.Drag.Finish (context, insert.Length > 0, false, time);
+				Gtk.Drag.Finish (context, more_than_one, false, time);
 			} else {
 				base.OnDragDataReceived (context, x, y, selection_data, info, time);
 			}
@@ -163,6 +178,12 @@ namespace Tomboy
 			
 			args.RetVal = ret_value;
 		}
+		
+		[GLib.ConnectBefore()]
+		void ButtonPressed (object sender, Gtk.ButtonPressEventArgs args)
+		{
+			((NoteBuffer) Buffer).CheckSelection ();
+		}
 	}
 
 	public class NoteWindow : ForcedPresentWindow 
@@ -176,6 +197,7 @@ namespace Tomboy
 		Gtk.Menu plugin_menu;
 		Gtk.TextView editor;
 		Gtk.ScrolledWindow editor_window;
+		NoteTagBar tag_bar;
 		NoteFindBar find_bar;
 
 		GlobalKeybinder global_keys;
@@ -224,7 +246,12 @@ namespace Tomboy
 
 			toolbar = MakeToolbar ();
 			toolbar.Show ();
-
+			
+			tag_bar = new NoteTagBar (note);
+			tag_bar.Visible = false;
+			tag_bar.NoShowAll = true;
+			tag_bar.Hidden += TagBarHidden;
+			
 			// The main editor widget
 			editor = new NoteEditor (note.Buffer);
 			editor.PopulatePopup += OnPopulatePopup;
@@ -253,10 +280,18 @@ namespace Tomboy
 
 			Gtk.VBox box = new Gtk.VBox (false, 2);
 			box.PackStart (toolbar, false, false, 0);
+			box.PackStart (tag_bar, false, false, 0);
 			box.PackStart (editor_window, true, true, 0);
 			box.PackStart (find_bar, false, false, 0);
 			box.Show ();
-
+			
+			// Don't set up Ctrl-W or Ctrl-N if Emacs is in use
+			bool using_emacs = false;
+			string gtk_key_theme = (string)
+					Preferences.Get ("/desktop/gnome/interface/gtk_key_theme");
+			if (gtk_key_theme != null && gtk_key_theme.CompareTo ("Emacs") == 0)
+				using_emacs = true;
+			
 			// NOTE: Since some of our keybindings are only
 			// available in the context menu, and the context menu
 			// is created on demand, register them with the
@@ -264,7 +299,8 @@ namespace Tomboy
 			global_keys = new GlobalKeybinder (accel_group);
 
 			// Close window (Ctrl-W)
-			global_keys.AddAccelerator (new EventHandler (CloseWindowHandler),
+			if (!using_emacs)
+				global_keys.AddAccelerator (new EventHandler (CloseWindowHandler),
 						    (uint) Gdk.Key.w, 
 						    Gdk.ModifierType.ControlMask,
 						    Gtk.AccelFlags.Visible);
@@ -298,7 +334,8 @@ namespace Tomboy
 						    0);
 			
 			// Create a new note
-			global_keys.AddAccelerator (new EventHandler (CreateNewNote),
+			if (!using_emacs)
+				global_keys.AddAccelerator (new EventHandler (CreateNewNote),
 							(uint) Gdk.Key.n,
 							Gdk.ModifierType.ControlMask,
 							Gtk.AccelFlags.Visible);
@@ -307,13 +344,13 @@ namespace Tomboy
 			KeyPressEvent += KeyPressed;
 						   
 			// Increase Indent			    
-			global_keys.AddAccelerator (new EventHandler (IncreaseDepthHandler),
+			global_keys.AddAccelerator (new EventHandler (ChangeDepthRightHandler),
 						    (uint) Gdk.Key.Right, 
 						    Gdk.ModifierType.Mod1Mask,
 						    Gtk.AccelFlags.Visible);
 						    
 			// Decrease Indent
-			global_keys.AddAccelerator (new EventHandler (DecreaseDepthHandler),
+			global_keys.AddAccelerator (new EventHandler (ChangeDepthLeftHandler),
 						    (uint) Gdk.Key.Left, 
 						    Gdk.ModifierType.Mod1Mask,
 						    Gtk.AccelFlags.Visible);			
@@ -578,6 +615,20 @@ namespace Tomboy
 			toolbar.AppendWidget (text_button, 
 					      Catalog.GetString ("Set properties of text"), 
 					      null);
+			
+			// FIXME: Use a proper icon for the Tags button
+			Gtk.Widget tags_button =
+				toolbar.AppendItem (
+					Catalog.GetString ("Tags"),
+					Catalog.GetString ("Add/Remove note tags"),
+					null,
+					new Gtk.Image (Gtk.Stock.SelectFont, toolbar.IconSize),
+					new Gtk.SignalFunc (TagBarClicked));
+			tags_button.AddAccelerator ("activate",
+						    accel_group,
+						    (uint) Gdk.Key.t, 
+						    Gdk.ModifierType.ControlMask,
+						    Gtk.AccelFlags.Visible);
 
 			ToolMenuButton plugin_button = 
 				new ToolMenuButton (toolbar, 
@@ -609,7 +660,7 @@ namespace Tomboy
 		//
 		// Plugin toolbar menu
 		//
-		// Prefixed with Open Plugins Folder action, the rest being
+		// This menu can be
 		// populated by individual plugins using
 		// NotePlugin.AddPluginMenuItem().
 		//
@@ -617,26 +668,9 @@ namespace Tomboy
 		Gtk.Menu MakePluginMenu ()
 		{
 			Gtk.Menu menu = new Gtk.Menu ();
-
-			Gtk.ImageMenuItem open;
-			open = new Gtk.ImageMenuItem (Catalog.GetString ("_Open Plugins Folder"));
-			open.Image = new Gtk.Image (Gtk.Stock.Open, Gtk.IconSize.Menu);
-			open.Activated += OnOpenPluginsFolderActivate;
-			open.Show ();
-			menu.Add (open);
-
-			Gtk.SeparatorMenuItem sep = new Gtk.SeparatorMenuItem ();
-			sep.Show ();
-			menu.Add (sep);
-
 			return menu;
 		}
 
-		void OnOpenPluginsFolderActivate (object sender, EventArgs args)
-		{
-			note.Manager.PluginManager.ShowPluginsDirectory ();
-		}
-
 		//
 		// Find context menu
 		//
@@ -694,7 +728,7 @@ namespace Tomboy
 
 			return menu;
 		}
-
+		
 		//
 		// Open the find dialog, passing any currently selected text
 		//
@@ -727,6 +761,13 @@ namespace Tomboy
 			Find.FindPreviousButton.Click ();
 		}
 		
+		void TagBarHidden (object sender, EventArgs args)
+		{
+			// Reposition the current focus back to the editor so the
+			// cursor will be ready for typing.
+			editor.GrabFocus ();
+		}
+		
 		void FindBarHidden (object sender, EventArgs args)
 		{
 			// Reposition the current focus back to the editor so the
@@ -803,15 +844,258 @@ namespace Tomboy
 			SearchButtonClicked ();
 		}
 
-		void IncreaseDepthHandler (object sender, EventArgs args)
+		void ChangeDepthRightHandler (object sender, EventArgs args)
 		{
-			((NoteBuffer)editor.Buffer).IncreaseCursorDepth();
+			((NoteBuffer)editor.Buffer).ChangeCursorDepthDirectional (true);
 		}
 		
-		void DecreaseDepthHandler (object sender, EventArgs args)
+		void ChangeDepthLeftHandler (object sender, EventArgs args)
 		{
-			((NoteBuffer)editor.Buffer).DecreaseCursorDepth();
-		}	
+			((NoteBuffer)editor.Buffer).ChangeCursorDepthDirectional (false);
+		}
+		
+		void TagBarClicked ()
+		{
+			if (tag_bar.Visible)
+				tag_bar.Hide ();
+			else
+				tag_bar.Show ();
+		}
+	}
+	
+	public class NoteTagBar : Gtk.HBox
+	{
+		private Note note;
+		
+		WrapBox tag_cloud;
+		Gtk.Entry tag_entry;
+		Gtk.Button add_tag_button;
+		
+		#region Constructors
+		public NoteTagBar (Note note) : base (false, 4)
+		{
+			this.note = note;
+			
+			BorderWidth = 2;
+
+			Gtk.Button button = new Gtk.Button ();
+			button.Image = new Gtk.Image (Gtk.Stock.Close, Gtk.IconSize.Menu);
+			button.Relief = Gtk.ReliefStyle.None;
+			button.Clicked += HideTagBar;
+			button.Show ();
+			PackStart (button, false, false, 4);
+
+			Gtk.Label label = new Gtk.Label (
+					string.Format ("<span weight=\"bold\">{0}</span>",
+						Catalog.GetString ("Tags:")));
+			label.Xalign = 0;
+			label.UseMarkup = true;
+			label.Show ();
+			
+			PackStart (label, false, false, 0);
+			
+			tag_cloud = new WrapBox ();
+			tag_cloud.Show ();
+			PackStart (tag_cloud, true, true, 0);
+			
+			foreach (Tag tag in note.Tags) {
+				button = MakeTagButton (tag);
+				button.Show ();
+				tag_cloud.Insert (button);
+			}
+			
+			Gtk.EntryCompletion entry_completion = new Gtk.EntryCompletion ();
+			entry_completion.InlineCompletion = true;
+			entry_completion.PopupSingleMatch = true;
+			entry_completion.Model = TagManager.Tags;
+			entry_completion.MatchFunc = TagEntryCompletionMatchFunc;
+			entry_completion.MatchSelected += TagEntryCompletionMatchSelected;
+			Gtk.CellRendererText crt = new Gtk.CellRendererText ();
+			entry_completion.PackStart (crt, true);
+			entry_completion.SetCellDataFunc (crt,
+					new Gtk.CellLayoutDataFunc (TagEntryCompletionDataFunc));
+			
+			tag_entry = new Gtk.Entry ();
+			tag_entry.Completion = entry_completion;
+			tag_entry.Activated += TagEntryActivated;
+			tag_entry.Changed += TagEntryChanged;
+			// Bind ESC to close the TagBar if it's open and has focus.
+			tag_entry.KeyPressEvent += KeyPressed;
+			tag_entry.Show ();
+			PackStart (tag_entry, false, false, 0);
+			
+			add_tag_button = new Gtk.Button (String.Empty);
+			add_tag_button.Image = new Gtk.Image (Gtk.Stock.Add, Gtk.IconSize.Menu);
+			add_tag_button.Sensitive = false;
+			add_tag_button.Clicked += AddTagButtonClicked;
+			add_tag_button.Show ();
+			PackStart (add_tag_button, false, false, 0);
+
+			note.TagAdded += TagAddedHandler;
+			note.TagRemoved += TagRemovedHandler;
+		}
+		#endregion
+		
+		#region Private Methods
+		Gtk.Button MakeTagButton (Tag tag)
+		{
+			TagButton button = new TagButton (tag);
+			button.Clicked += TagButtonClicked;
+
+			return button;
+		}
+
+		protected override void OnShown ()
+		{
+			tag_entry.GrabFocus ();
+
+			base.OnShown ();
+		}
+		
+		protected override void OnHidden ()
+		{
+			base.OnHidden ();
+		}
+		
+		#endregion
+
+		#region Event Handlers
+		// <summary>
+		// When the user presses return, call the same method that is called
+		// when the user presses the Add button.
+		// </summary>
+		void TagEntryActivated (object sender, EventArgs args)
+		{
+			AddTagButtonClicked (sender, args);
+		}
+		
+		void TagEntryChanged (object sender, EventArgs args)
+		{
+			string text = tag_entry.Text.Trim ();
+			if (text.Length == 0) {
+				add_tag_button.Sensitive = false;
+			} else {
+				add_tag_button.Sensitive = true;
+			}
+		}
+		
+		void AddTagButtonClicked (object sender, EventArgs args)
+		{
+			string text = tag_entry.Text.Trim ();
+			if (text.Length > 0) {
+				Tag tag = TagManager.GetOrCreateTag (text);
+				if (!note.Data.Tags.ContainsKey (tag.NormalizedName)) {
+					note.AddTag (tag);
+				}
+			}
+			
+			// Clear out the entry and grab the cursor focus so it's ready to
+			// have another tag typed and added.
+			tag_entry.Text = String.Empty;
+		}
+		
+		void TagButtonClicked (object sender, EventArgs args)
+		{
+			TagButton button = sender as TagButton;
+			
+			note.RemoveTag (button.Tag);
+		}
+		
+		void TagAddedHandler (Note note, Tag tag)
+		{
+			Gtk.Button button = MakeTagButton (tag);
+			button.Show ();
+			tag_cloud.Insert (button);
+		}
+		
+		void TagRemovedHandler (Note note, string tag_name)
+		{
+			// Remove the button from the list
+			TagButton button_to_remove = null;
+			
+			foreach (Gtk.Widget w in tag_cloud.Children) {
+				if (w is TagButton) {
+					TagButton button = w as TagButton;
+					
+					if (string.Compare (button.Tag.NormalizedName, tag_name) == 0) {
+						button_to_remove = button;
+						button.Hide ();
+						break;
+					}
+				}
+			}
+			
+			if (button_to_remove != null)
+				tag_cloud.Remove (button_to_remove);
+			else
+				Logger.Debug ("\tDid not remove a tag button");
+		}
+		
+		void TagEntryCompletionDataFunc (Gtk.CellLayout cell_layout, Gtk.CellRenderer cell,
+				Gtk.TreeModel tree_model, Gtk.TreeIter iter)
+		{
+			Gtk.CellRendererText crt = cell as Gtk.CellRendererText;
+			Tag tag = tree_model.GetValue (iter, 0) as Tag;
+			crt.Text = tag.Name;
+		}
+		
+		bool TagEntryCompletionMatchFunc (Gtk.EntryCompletion completion, string key, Gtk.TreeIter iter)
+		{
+			if (key.Length == 0)
+				return true;
+			
+			Tag tag = completion.Model.GetValue (iter, 0) as Tag;
+			if (tag == null)
+				return false;
+			
+			string lower_cased_key = key.ToLower ();
+			
+			if (tag.NormalizedName.StartsWith (lower_cased_key))
+				return true;
+			
+			return false;
+		}
+		
+		[GLib.ConnectBefore]
+		void TagEntryCompletionMatchSelected (object sender, Gtk.MatchSelectedArgs args)
+		{
+			Tag tag = args.Model.GetValue (args.Iter, 0) as Tag;
+			if (tag != null) {
+				tag_entry.Text = tag.Name;
+				args.RetVal = true;
+				
+				// Press the add button for the user
+				AddTagButtonClicked (sender, EventArgs.Empty);
+			}
+		}
+		
+		void HideTagBar (object sender, EventArgs args)
+		{
+			Hide ();
+		}
+		
+		void KeyPressed (object sender, Gtk.KeyPressEventArgs args)
+		{
+			args.RetVal = true;
+			
+			switch (args.Event.Key)
+			{
+			case Gdk.Key.Escape:
+				Hide ();
+				break;
+			default:
+				args.RetVal = false;
+				break;
+			}
+		}
+
+		#endregion
+
+		#region Public Methods
+		#endregion
+		
+		#region Properties
+		#endregion
 	}
 	
 	public class NoteFindBar : Gtk.HBox
@@ -827,6 +1111,7 @@ namespace Tomboy
 		string prev_search_text;
 		
 		InterruptableTimeout entry_changed_timeout;
+		InterruptableTimeout note_changed_timeout;
 
 		bool shift_key_pressed;
 		
@@ -910,7 +1195,7 @@ namespace Tomboy
 			note.Buffer.InsertText -= OnInsertText;
 			note.Buffer.DeleteRange -= OnDeleteRange;
 
-			base.OnShown ();
+			base.OnHidden ();
 		}
 		
 		void HideFindBar (object sender, EventArgs args)
@@ -979,20 +1264,22 @@ namespace Tomboy
 
 		void OnCaseSensitiveToggled (object sender, EventArgs args)
 		{
-			PerformSearch ();
+			PerformSearch (true);
 		}
 		
 		void OnFindEntryActivated (object sender, EventArgs args)
 		{
-			if (entry_changed_timeout != null)
+			if (entry_changed_timeout != null) {
 				entry_changed_timeout.Cancel ();
+				entry_changed_timeout = null;
+			}
 			
 			if (prev_search_text != null && 
 			    SearchText != null && 
 			    prev_search_text.CompareTo (SearchText) == 0)
 				next_button.Click ();
 			else
-				PerformSearch ();
+				PerformSearch (true);
 		}
 		
 		void OnFindEntryChanged (object sender, EventArgs args)
@@ -1003,23 +1290,25 @@ namespace Tomboy
 			}
 			
 			if (SearchText == null) {
-				PerformSearch ();
+				PerformSearch (false);
 			} else {
 				entry_changed_timeout.Reset (500);
 			}
 		}
 		
 		// Called after .5 seconds of typing inactivity, or on explicit
-		// activate.  Redo the serach and update the results...
+		// activate.  Redo the search and update the results...
 		void EntryChangedTimeout (object sender, EventArgs args)
 		{
+			entry_changed_timeout = null;
+			
 			if (SearchText == null)
 				return;
 			
-			PerformSearch ();
+			PerformSearch (true);
 		}
 		
-		void PerformSearch ()
+		void PerformSearch (bool scroll_to_hit)
 		{
 			CleanupMatches ();
 
@@ -1043,7 +1332,8 @@ namespace Tomboy
 				HighlightMatches (true);
 				
 				// Select/scroll to the first match
-				OnNextClicked (this, EventArgs.Empty);
+				if (scroll_to_hit)
+					OnNextClicked (this, EventArgs.Empty);
 			}
 
 			UpdateSensitivity ();
@@ -1065,14 +1355,42 @@ namespace Tomboy
 			}
 		}
 		
+		void UpdateSearch ()
+		{
+			if (note_changed_timeout == null) {
+				note_changed_timeout = new InterruptableTimeout ();
+				note_changed_timeout.Timeout += NoteChangedTimeout;
+			}
+			
+			if (SearchText == null) {
+				PerformSearch (false);
+			} else {
+				note_changed_timeout.Reset (500);
+			}
+		}
+		
+		// Called after .5 seconds of typing inactivity to update
+		// the search when the text of a note changes.  This prevents
+		// the search from running on every single change made in a
+		// note.
+		void NoteChangedTimeout (object sender, EventArgs args)
+		{
+			note_changed_timeout = null;
+			
+			if (SearchText == null)
+				return;
+			
+			PerformSearch (false);
+		}
+		
 		void OnInsertText (object sender, Gtk.InsertTextArgs args)
 		{
-			PerformSearch ();
+			UpdateSearch ();
 		}
 		
 		void OnDeleteRange (object sender, Gtk.DeleteRangeArgs args)
 		{
-			PerformSearch ();
+			UpdateSearch ();
 		}
 		
 		//
@@ -1371,7 +1689,7 @@ namespace Tomboy
 			normal = new Gtk.RadioMenuItem (Catalog.GetString ("_Normal"));
 			MarkupLabel (normal);
 			normal.Active = true;
-			normal.Toggled += FontSizeClicked;
+			normal.Activated += FontSizeActivated;
 
 			huge = new Gtk.RadioMenuItem (normal.Group, 
 						      "<span size=\"x-large\">" +
@@ -1379,7 +1697,7 @@ namespace Tomboy
 						      "</span>");
 			MarkupLabel (huge);
 			huge.Data ["Tag"] = "size:huge";
-			huge.Toggled += FontSizeClicked;
+			huge.Activated += FontSizeActivated;
 
 			large = new Gtk.RadioMenuItem (huge.Group, 
 						       "<span size=\"large\">" +
@@ -1387,7 +1705,7 @@ namespace Tomboy
 						       "</span>");
 			MarkupLabel (large);
 			large.Data ["Tag"] = "size:large";
-			large.Toggled += FontSizeClicked;
+			large.Activated += FontSizeActivated;
 
 			small = new Gtk.RadioMenuItem (large.Group, 
 						       "<span size=\"small\">" +
@@ -1395,7 +1713,7 @@ namespace Tomboy
 						       "</span>");
 			MarkupLabel (small);
 			small.Data ["Tag"] = "size:small";
-			small.Toggled += FontSizeClicked;
+			small.Activated += FontSizeActivated;
 
 			hidden_no_size = new Gtk.RadioMenuItem (small.Group, string.Empty);
 			hidden_no_size.Hide ();
@@ -1475,9 +1793,11 @@ namespace Tomboy
 			highlight.Active = buffer.IsActiveTag ("highlight");
 			
 			bool inside_bullets = buffer.IsBulletedListActive ();
+			bool can_make_bulleted_list = buffer.CanMakeBulletedList ();
 			bullets.Activated -= ToggleBulletsClicked;
 			bullets.Active = inside_bullets;
 			bullets.Activated += ToggleBulletsClicked;
+			bullets.Sensitive = can_make_bulleted_list;
 			increase_indent.Sensitive = inside_bullets;
 			decrease_indent.Sensitive = inside_bullets;
 
@@ -1516,8 +1836,13 @@ namespace Tomboy
 		// Set the font size tag for the current text.  Style tags are
 		// stored in a "Tag" member of the menuitem's Data.
 		//
-
-		void FontSizeClicked (object sender, EventArgs args) 
+		
+		// FIXME: Change this back to use FontSizeToggled instead of using the
+		// Activated signal.  Fix the Text menu so it doesn't show a specific
+		// font size already selected if multiple sizes are highlighted. The
+		// Activated event is used here to fix
+		// http://bugzilla.gnome.org/show_bug.cgi?id=412404.
+		void FontSizeActivated (object sender, EventArgs args)
 		{
 			if (event_freeze)
 				return;
@@ -1562,20 +1887,7 @@ namespace Tomboy
 		//
 		void ToggleBulletsClicked (object sender, EventArgs args)
 		{
-			bool inside_bullets = buffer.IsBulletedListActive ();
-			if (!inside_bullets)
-				buffer.IncreaseCursorDepth ();
-			else {
-				// FIXME: This is a bad hack.  There's got to be
-				// a better way to remove the bullet.
-				
-				// For now, keep calling buffer.DecreaseCursorDepth ()
-				// until bullets are no longer present.
-				do {
-					buffer.DecreaseCursorDepth ();
-					inside_bullets = buffer.IsBulletedListActive ();
-				} while (inside_bullets == true);
-			}
+			buffer.ToggleSelectionBullets ();
 		}
 		
 		void IncreaseIndentClicked (object sender, EventArgs args)
