@@ -41,6 +41,27 @@ namespace Tomboy
 
 		Type preferencesWidget;
 
+		public const string OFFICIAL_AUTHOR = "official";
+		
+		// The default constructor is, for some reason, needed or Tomboy will
+		// crash when attempting to read the plugin attributes.
+		public PluginInfoAttribute ()
+		{
+			// Intentionally blank
+		}
+
+		public PluginInfoAttribute (string name, string version,
+		                            string author, string description)
+		{
+			this.name = Catalog.GetString (name);
+			this.description = Catalog.GetString (description);
+			this.version = version;
+			if (author == OFFICIAL_AUTHOR)
+				this.author = Catalog.GetString ("Tomboy Project");
+			else
+				this.author = author;
+		}
+
 		public string Name
 		{
 			get { return name; }
@@ -302,6 +323,7 @@ namespace Tomboy
 			typeof (NoteLinkWatcher),
 			typeof (NoteWikiWatcher),
 			typeof (MouseHandWatcher),
+			typeof (NoteTagsWatcher),
 			
 			// Not ready yet:
 			// typeof (NoteRelatedToWatcher),
@@ -438,29 +460,6 @@ namespace Tomboy
 
 			// Clean up old plugin remnants
 			CleanupOldPlugins (plugins_dir);
-
-			// Copy Plugins/DefaultPlugins.desktop file from resource
-			string default_desktop = "DefaultPlugins.desktop";
-			string default_path = Path.Combine (plugins_dir, default_desktop);
-			if (!File.Exists (default_path)) {
-				Assembly asm = Assembly.GetExecutingAssembly();
-				Stream stream = asm.GetManifestResourceStream (default_desktop);
-				if (stream != null) {
-					Logger.Log ("Writing '{0}'...", default_path);
-					try {
-						StreamWriter file = File.CreateText (default_path);
-						StreamReader reader;
-						try {
-							reader = new StreamReader (stream);
-							file.Write (reader.ReadToEnd ());
-						} finally {
-							file.Close ();
-						}
-					} finally {
-						stream.Close ();
-					}
-				}
-			}
 		}
 
 		void OnNoteDeleted (object sender, Note deleted)
@@ -515,8 +514,13 @@ namespace Tomboy
 		void AttachPlugin (Type type)
 		{
 			if (typeof (NotePlugin).IsAssignableFrom (type)) {
-				foreach (Note note in attached_plugins.Keys as
-						System.Collections.Generic.IEnumerable<Note>)
+				// A plugin may add or remove notes when being
+				// created. Therefore, it is best to iterate
+				// through a copy of the notes list to avoid
+				// "System.InvalidOperationException: out of sync"
+				List<Note> notes_copy =
+					new List<Note> (attached_plugins.Keys);				
+				foreach (Note note in notes_copy)
 					AttachPlugin (type, note);
 			}
 		}
